﻿B4J=true
Group=Default Group
ModulesStructureVersion=1
Type=Class
Version=6.3
@EndOfDesignText@
Sub Class_Globals
	Private xui As XUI 'ignore
	Type BCEPixelGroup (SrcX As Int, SrcY As Int, x As Float, y As Float, dx As Float, dy As Float)
	Type HSVColor (H As Int, S As Float, V As Float, A As Float)
	Private SleepDuration As Int
	Public ScaleDownImages As Boolean = True
End Sub

Public Sub Initialize
	If xui.IsB4i Then
		SleepDuration = 5
	Else
		SleepDuration = 16
	End If
End Sub

'Returns 1 in B4J and B4A.
'Returns the non-normalized scale in B4i when ScaleDownImages is set to False.
Public Sub getImageScale As Float
	#if B4i
	If ScaleDownImages Then Return 1
	Return GetDeviceLayoutValues.NonnormalizedScale
	#else
	Return 1
	#End If
End Sub

Public Sub CreateBC(bmp As B4XBitmap) As BitmapCreator
	Dim bc As BitmapCreator = CreateEmptyBC(bmp)
	bc.CopyPixelsFromBitmap(bmp)
	Return bc
End Sub

'Creates an empty BitmapCreator with the size based on the input bitmap.
Public Sub CreateEmptyBC (bmp As B4XBitmap) As BitmapCreator
	Dim bc As BitmapCreator
	If ScaleDownImages And xui.IsB4A Then
		bc.Initialize(bmp.Width / bmp.Scale, bmp.Height / bmp.Scale)
	Else
		bc.Initialize(bmp.Width, bmp.Height)
	End If
	Return bc
End Sub

Public Sub GreyScale (bmp As B4XBitmap) As B4XBitmap
	Dim bc As BitmapCreator = CreateBC(bmp)
	Dim argb As ARGBColor
	For x = 0 To bc.mWidth - 1
		For y = 0 To bc.mHeight - 1
			bc.GetARGB(x, y, argb)
			Dim c As Int = argb.r * 0.21 + argb.g * 0.72 + 0.07 * argb.b
			argb.r = c
			argb.g = c
			argb.b = c
			bc.SetARGB(x, y, argb)
		Next
	Next
	Return bc.Bitmap
End Sub

Public Sub Negate (bmp As B4XBitmap) As B4XBitmap
	Dim bc As BitmapCreator = CreateBC(bmp)
	Dim argb As ARGBColor
	For x = 0 To bc.mWidth - 1
		For y = 0 To bc.mHeight - 1
			bc.GetARGB(x, y, argb)
			argb.r = Bit.Xor(argb.r, 0xff)
			argb.g = Bit.Xor(argb.g, 0xff)
			argb.b = Bit.Xor(argb.b, 0xff)
			bc.SetARGB(x, y, argb)
		Next
	Next
	Return bc.Bitmap
End Sub

Public Sub Blur (bmp As B4XBitmap) As B4XBitmap
	Dim bc As BitmapCreator
	Dim ReduceScale As Int = 2
	bc.Initialize(bmp.Width / ReduceScale / bmp.Scale, bmp.Height / ReduceScale / bmp.Scale)
	bc.CopyPixelsFromBitmap(bmp)
	Dim count As Int = 3
	Dim clrs(3) As ARGBColor
	Dim temp As ARGBColor
	Dim m As Int
	For steps = 1 To count
		For y = 0 To bc.mHeight - 1
			For x = 0 To 2
				bc.GetARGB(x, y, clrs(x))
			Next
			SetAvg(bc, 1, y, clrs, temp)
			m = 0
			For x = 2 To bc.mWidth - 2
				bc.GetARGB(x + 1, y, clrs(m))
				m = (m + 1) Mod clrs.Length
				SetAvg(bc, x, y, clrs, temp)
			Next
		Next
		For x = 0 To bc.mWidth - 1
			For y = 0 To 2
				bc.GetARGB(x, y, clrs(y))
			Next
			SetAvg(bc, x, 1, clrs, temp)
			m = 0
			For y = 2 To bc.mHeight - 2
				bc.GetARGB(x, y + 1, clrs(m))
				m = (m + 1) Mod clrs.Length
				SetAvg(bc, x, y, clrs, temp)
			Next
		Next
	Next
	Dim res As B4XBitmap = bc.Bitmap
	If ReduceScale > 1 Then res = res.Resize(ReduceScale * res.Width * xui.Scale, ReduceScale * res.Height * xui.Scale, False)
	Return res
End Sub

Private Sub SetAvg(bc As BitmapCreator, x As Int, y As Int, clrs() As ARGBColor, temp As ARGBColor)
	temp.Initialize
	For Each c As ARGBColor In clrs
		temp.r = temp.r + c.r
		temp.g = temp.g + c.g
		temp.b = temp.b + c.b
	Next
	temp.a = 255
	temp.r = temp.r / clrs.Length
	temp.g = temp.g / clrs.Length
	temp.b = temp.b / clrs.Length
	bc.SetARGB(x, y, temp)
End Sub

Public Sub Brightness(Bmp As B4XBitmap, Factor As Float) As B4XBitmap
	Dim bc As BitmapCreator = CreateBC(Bmp)
	Dim argb As ARGBColor
	For x = 0 To bc.mWidth - 1
		For y = 0 To bc.mHeight - 1
			bc.GetARGB(x, y, argb)
			argb.r = Min(255, argb.r * Factor)
			argb.g = Min(255, argb.g * Factor)
			argb.b = Min(255, argb.b * Factor)
			bc.SetARGB(x, y, argb)
		Next
	Next
	Return bc.Bitmap
End Sub

Public Sub Pixelate (Bmp As B4XBitmap, BoxSize As Int) As B4XBitmap
	Dim bc As BitmapCreator = CreateBC(Bmp)
	Dim rect As B4XRect
	rect.Initialize(0, 0, 0, 0)
	For x = 0 To bc.mWidth - 1 Step BoxSize
		rect.Left = x
		rect.Width = BoxSize
		For y = 0 To bc.mHeight - 1 Step BoxSize
			rect.Top = y
			rect.Height = BoxSize
			bc.FillRect(bc.GetColor(Min(bc.mWidth - 1, x + BoxSize / 2), _
				 Min(bc.mHeight - 1, y + BoxSize / 2)), rect)
		Next
	Next
	Return bc.Bitmap
End Sub

Public Sub PixelateAnimated (Duration As Int, Bmp As B4XBitmap, FromBoxSize As Int, ToBoxSize As Int, ImageView As B4XView) As ResumableSub
	Dim steps As Int = Min(10, Abs(ToBoxSize - FromBoxSize))
	Dim SleepLength As Int = Duration / steps
	Dim delta As Int = Round((ToBoxSize - FromBoxSize) / steps)
	For i = FromBoxSize To ToBoxSize Step delta
		SetBitmap(ImageView, Pixelate(Bmp, i))
		Sleep(SleepLength)
	Next
	SetBitmap(ImageView, Pixelate(Bmp, ToBoxSize))
	Return True
End Sub

'assuming that mask is the same size or larger than Bmp.
Public Sub DrawThroughMask (Bmp As B4XBitmap, Mask As B4XBitmap) As B4XBitmap
	Dim source As BitmapCreator = CreateBC(Bmp)
	Dim maskbc As BitmapCreator = CreateBC(Mask)
	Dim transparent As ARGBColor
	transparent.a = 0
	Dim argb1, argb2 As ARGBColor
	For x = 0 To source.mWidth - 1
		For y = 0 To source.mHeight - 1
			If maskbc.IsTransparent(x, y) Then
				source.SetARGB(x, y, transparent)
			Else
				maskbc.GetARGB(x, y, argb1)
				If argb1.a < 255 Then
					source.GetARGB(x, y, argb2)
					argb2.a = (argb2.a * argb1.a) / 255
					source.SetARGB(x, y, argb2)
				End If
			End If
		Next
	Next
	Return source.Bitmap
End Sub

'assuming that mask is the same size or larger than Bmp.
Public Sub DrawOutsideMask (Bmp As B4XBitmap, Mask As B4XBitmap) As B4XBitmap
	Dim source As BitmapCreator = CreateBC(Bmp)
	Dim maskbc As BitmapCreator = CreateBC(Mask)
	Dim transparent As ARGBColor
	transparent.a = 0
	Dim argb1, argb2 As ARGBColor
	For x = 0 To source.mWidth - 1
		For y = 0 To source.mHeight - 1
			If maskbc.IsTransparent(x, y) Then
				
			Else
				maskbc.GetARGB(x, y, argb1)
				If argb1.a = 255 Then
					source.SetARGB(x, y, transparent)
				Else
					source.GetARGB(x, y, argb2)
					argb2.a = (argb2.a * (255 - argb1.a)) / 255
					source.SetARGB(x, y, argb2)
				End If
			End If
		Next
	Next
	Return source.Bitmap
End Sub

Public Sub ImplodeAnimated (Duration As Int, Bmp As B4XBitmap, ImageView As B4XView, PieceSize As Int) As ResumableSub
	Dim source As BitmapCreator = CreateBC(Bmp)
	
	Dim NumberOfSteps As Int = Duration / 20
	Dim steps As Int = NumberOfSteps
	Dim GroupSize As Int = PieceSize
	Dim w As Int = Bmp.Width / Bmp.Scale / GroupSize
	Dim h As Int = Bmp.Height / Bmp.Scale / GroupSize
	Dim pgs(w, h) As BCEPixelGroup
	Dim target As BitmapCreator
	target.Initialize(w * GroupSize, h * GroupSize)
	For x = 0 To w - 1
		For y = 0 To h - 1
			Dim pg As BCEPixelGroup = pgs(x, y)
			pg.SrcX = x * GroupSize
			pg.Srcy = y * GroupSize
			Select Rnd(0, 4)
				Case 0
					pg.x = 0
					pg.y = Rnd(0, target.mHeight)
				Case 1
					pg.x = target.mWidth - 1
					pg.y = Rnd(0, target.mHeight)
				Case 2
					pg.x = Rnd(0, target.mWidth)
					pg.y = 0
				Case 3
					pg.x = Rnd(0, target.mWidth)
					pg.y = target.mHeight - 1
			End Select
			
			pg.dx = (pg.SrcX - pg.x) / steps
			pg.dy = (pg.SrcY - pg.y) / steps
		Next
	Next
	Dim r As B4XRect
	r.Initialize(0, 0, 0, 0)
	For i = 0 To steps - 1
		target.FillRect(xui.Color_Transparent, target.TargetRect)
		For x = 0 To w - 1
			For y = 0 To h - 1
				Dim pg As BCEPixelGroup = pgs(x, y)
				pg.x = pg.x + pg.dx
				pg.y = pg.y + pg.dy
				r.Left = pg.SrcX
				r.Right = pg.SrcX + GroupSize
				r.Top = pg.SrcY
				r.Bottom = pg.SrcY + GroupSize
				target.DrawBitmapCreator(source, r, pg.x, pg.y, True)
			Next
		Next
		SetBitmap(ImageView, target.Bitmap)
		Sleep(SleepDuration)
	Next
	SetBitmap(ImageView, Bmp)
	Return True
End Sub

'Replaces OldColor with NewColor.
'KeepAlphaLevel - If true then the alpha level of the source color is kept.
Public Sub ReplaceColor (bmp As B4XBitmap, OldColor As Int, NewColor As Int, KeepAlphaLevel As Boolean) As B4XBitmap
	Dim bc As BitmapCreator = CreateBC(bmp)
	Dim oldargb, newargb, a As ARGBColor
	bc.ColorToARGB(OldColor, oldargb)
	bc.ColorToARGB(NewColor, newargb)
	For x = 0 To bc.mWidth - 1
		For y = 0 To bc.mHeight - 1
			bc.GetARGB(x, y, a)
			If (KeepAlphaLevel Or a.a = oldargb.a) And a.r = oldargb.r And a.g = oldargb.g And a.b = oldargb.b Then
				newargb.a = a.a
				bc.SetARGB(x, y, newargb)
			End If
		Next
	Next
	Return bc.Bitmap
End Sub

Public Sub FlipHorizontal (bmp As B4XBitmap) As B4XBitmap
	Dim bc As BitmapCreator = CreateBC(bmp)
	Dim c As Float = (bc.mWidth - 1) / 2
	Dim argbLeft, argbRight As ARGBColor
	For x = 0 To c - 0.5
		For y = 0 To bc.mHeight - 1
			bc.GetARGB(x, y, argbLeft)
			bc.GetARGB(bc.mWidth -1 - x, y, argbRight)
			bc.SetARGB(x, y,argbRight)
			bc.SetARGB(bc.mWidth -1 - x, y, argbLeft)
		Next
	Next
	Return bc.Bitmap
End Sub

Public Sub FlipVertical (bmp As B4XBitmap) As B4XBitmap
	Dim bc As BitmapCreator = CreateBC(bmp)
	Dim c As Float = (bc.mHeight - 1) / 2
	Dim argbTop, argbDown As ARGBColor
	For x = 0 To bc.mWidth - 1
		For y = 0 To c
			bc.GetARGB(x, y, argbTop)
			bc.GetARGB(x, bc.mHeight - 1 - y, argbDown)
			bc.SetARGB(x, y,argbDown)
			bc.SetARGB(x, bc.mHeight - 1 - y, argbTop)
		Next
	Next
	Return bc.Bitmap
End Sub

Private Sub FadePixel(bc As BitmapCreator, argb As ARGBColor, x As Int, y As Int, f As Float)
	bc.GetARGB(x, y, argb)
	argb.a = argb.a * f
	bc.SetARGB(x, y, argb)
End Sub

Public Sub FadeBorders (bmp As B4XBitmap, Width As Int) As B4XBitmap
	Dim bc As BitmapCreator = CreateBC(bmp)
	Dim argb As ARGBColor
	For y = 0 To bc.mHeight - 1
		For x = 0 To Width - 1
			Dim f As Float = (x + 1) / Width
			FadePixel(bc, argb, x, y, f)
			FadePixel(bc, argb, bc.mWidth - 1 - x, y, f)
		Next
	Next
	For x = 0 To bc.mWidth - 1
		For y = 0 To Width - 1
			Dim f As Float = (y + 1) / Width
			FadePixel(bc, argb, x, y, f)
			FadePixel(bc, argb, x, bc.mHeight - 1 - y, f)
		Next
	Next
	
	Return bc.Bitmap
End Sub

Public Sub TransitionAnimated (Duration As Int, FromBmp As B4XBitmap, ToBmp As B4XBitmap, ImageView As B4XView) As ResumableSub
	Dim frombc As BitmapCreator = CreateBC(FromBmp)
	Dim tobc As BitmapCreator = CreateBC(ToBmp)
	Dim target As BitmapCreator
	target.Initialize(frombc.mWidth, frombc.mHeight)
	Dim StartTime As Long = DateTime.Now
	Dim EndTime As Long = StartTime + Duration
	Dim fromclr, toclr As ARGBColor
	Do While DateTime.Now < EndTime
		Dim progress As Float = (DateTime.Now - StartTime) / Duration
		For x = 0 To frombc.mWidth - 1
			For y = 0 To frombc.mHeight - 1
				frombc.GetARGB(x, y, fromclr)
				tobc.GetARGB(x, y, toclr)
				toclr.a = fromclr.a + progress * (toclr.a - fromclr.a)
				toclr.r = fromclr.r + progress * (toclr.r - fromclr.r)
				toclr.g = fromclr.g + progress * (toclr.g - fromclr.g)
				toclr.b = fromclr.b + progress * (toclr.b - fromclr.b)
				target.SetARGB(x, y, toclr)
			Next
		Next
		SetBitmap(ImageView, target.Bitmap)
		Sleep(SleepDuration)
	Loop
	SetBitmap(ImageView, target.Bitmap)
	Return True
End Sub

Private Sub SetBitmap(iv As B4XView, bmp As B4XBitmap)
	If iv.Tag Is B4XImageView Then
		Dim v As B4XImageView = iv.Tag
		v.Bitmap = bmp
	Else
		iv.SetBitmap(bmp)
	End If
End Sub

Public Sub AdjustColors(Bmp As B4XBitmap, HueOffset As Int, SaturationFactor As Float) As B4XBitmap
	Dim bc As BitmapCreator = CreateBC(Bmp)
	Dim hsv As HSVColor
	Dim argb As ARGBColor
	For y = 0 To bc.mHeight - 1
		For x = 0 To bc.mWidth - 1
			bc.GetARGB(x, y, argb)
			ARGBToHSV(argb, hsv)
			hsv.S = hsv.S * SaturationFactor
			hsv.H = (hsv.H + HueOffset) Mod 360
			bc.SetHSV(x, y, hsv.A, hsv.H, hsv.S, hsv.V)
		Next
	Next
	Return bc.Bitmap
End Sub

Private Sub ARGBToHSV(argb As ARGBColor, HSV As HSVColor)
	Dim a As Int = argb.a
	Dim r As Int = argb.r
	Dim g As Int = argb.g
	Dim b As Int = argb.b
	Dim h, s, v As Float
	Dim cmax As Int = Max(Max(r, g), b)
	Dim cmin As Int = Min(Min(r, g), b)
	v = cmax / 255
	If cmax <> 0 Then
		s = (cmax - cmin) / cmax
	End If
	If s = 0 Then
		h = 0
	Else
		Dim rc As Float = (cmax - r) / (cmax - cmin)
		Dim gc As Float = (cmax - g) / (cmax - cmin)
		Dim bc As Float = (cmax - b) / (cmax - cmin)
		If r = cmax Then
			h = bc - gc
		Else If g = cmax Then
			h = 2 + rc - bc
		Else
			h = 4 + gc - rc
		End If
		h = h / 6
		If h < 0 Then h = h + 1
	End If
	HSV.H = h * 360
	HSV.S = s
	HSV.V = v
	HSV.A = a
End Sub

'*******************************************************************
'* New subs from here on
'*******************************************************************

public Sub ApplyTint(bmp As B4XBitmap, tintColor As Int) As B4XBitmap
	
	Dim bc As BitmapCreator = CreateBC(bmp)
	Dim red As Float = Bit.And(Bit.ShiftRight(tintColor, 16), 0xFF) / 255
	Dim green As Float = Bit.And(Bit.ShiftRight(tintColor, 8), 0xFF) / 255
	Dim blue As Float = Bit.And(tintColor, 0xFF) / 255
	Dim alpha As Float = Bit.And(Bit.ShiftRight(tintColor, 24), 0xFF) / 255

	For y = 0 To bc.mHeight - 1
		For x = 0 To bc.mWidth - 1
			Dim argb As ARGBColor 
			bc.GetARGB (x, y, argb)
			argb.r = argb.r * (1 - alpha) + red * alpha * 255
			argb.g = argb.g * (1 - alpha) + green * alpha * 255
			argb.b = argb.b * (1 - alpha) + blue * alpha * 255
			bc.SetARGB(x, y, argb)
		Next
	Next
	
	Return bc.Bitmap
End Sub

Public Sub AdjustTemperature(bmp As B4XBitmap, adjustment As Float) As B4XBitmap
	Dim bc As BitmapCreator = CreateBC(bmp)
	For y = 0 To bc.mHeight - 1
		For x = 0 To bc.mWidth - 1
			Dim argb As ARGBColor
			bc.GetARGB(x, y, argb)
            
			' Adjust the red and blue channels based on the temperature adjustment
			argb.r = AdjustRedChannel(argb.r, adjustment)
			argb.b = AdjustBlueChannel(argb.b, adjustment)
            
			bc.SetARGB(x, y, argb)
		Next
	Next
	Return bc.Bitmap
End Sub

Private Sub AdjustRedChannel(color As Int, adjustment As Float) As Int
	' Gradual increase of the red channel based on adjustment
	Dim newColor As Float = color + (adjustment * (255 - color))
	Return Min(255, Max(0, Round(newColor)))
End Sub

Private Sub AdjustBlueChannel(color As Int, adjustment As Float) As Int
	' Gradual decrease of the blue channel based on adjustment
	Dim newColor As Float = color - (adjustment * color)
	Return Min(255, Max(0, Round(newColor)))
End Sub

Public Sub AdjustExposure(bmp As B4XBitmap, adjustment As Float) As B4XBitmap
	Dim bc As BitmapCreator = CreateBC(bmp)
	For y = 0 To bc.mHeight - 1
		For x = 0 To bc.mWidth - 1
			Dim argb As ARGBColor
			bc.GetARGB(x, y, argb)
            
			' Apply exposure adjustment using gamma correction-like method
			argb.r = AdjustColor(argb.r, adjustment)
			argb.g = AdjustColor(argb.g, adjustment)
			argb.b = AdjustColor(argb.b, adjustment)
            
			bc.SetARGB(x, y, argb)
		Next
	Next
	Return bc.Bitmap
End Sub

Private Sub AdjustColor(color As Int, adjustment As Float) As Int
	Dim newColor As Float = 255 * Power(color / 255.0, 1 / adjustment)
	Return Min(255, Max(0, Round(newColor)))
End Sub

public Sub AdjustContrast(bmp As B4XBitmap, adjustment As Float) As B4XBitmap
	Dim bc As BitmapCreator = CreateBC(bmp)
	Dim factor As Float = (259 * (adjustment + 255)) / (255 * (259 - adjustment))
    
	For y = 0 To bc.mHeight - 1
		For x = 0 To bc.mWidth - 1
			Dim argb As ARGBColor
			bc.GetARGB(x, y, argb)
			argb.r = Min(255, Max(0, factor * (argb.r - 128) + 128))
			argb.g = Min(255, Max(0, factor * (argb.g - 128) + 128))
			argb.b = Min(255, Max(0, factor * (argb.b - 128) + 128))
			bc.SetARGB(x, y, argb)
		Next
	Next
	Return bc.Bitmap
End Sub

public Sub AdjustHighlights(bmp As B4XBitmap, adjustment As Float) As B4XBitmap
	Dim bc As BitmapCreator = CreateBC(bmp)
	For y = 0 To bc.mHeight - 1
		For x = 0 To bc.mWidth - 1
			Dim argb As ARGBColor
			bc.GetARGB(x, y, argb)
			' Adjust highlights by modifying the brighter pixels
			If argb.r > 128 Then argb.r = Min(255, Max(0, 128 + (argb.r - 128) * adjustment))
			If argb.g > 128 Then argb.g = Min(255, Max(0, 128 + (argb.g - 128) * adjustment))
			If argb.b > 128 Then argb.b = Min(255, Max(0, 128 + (argb.b - 128) * adjustment))
			bc.SetARGB(x, y, argb)
		Next
	Next
	Return bc.Bitmap
End Sub

public Sub AdjustShadows(bmp As B4XBitmap, adjustment As Float) As B4XBitmap
	Dim bc As BitmapCreator = CreateBC(bmp)
	For y = 0 To bc.mHeight - 1
		For x = 0 To bc.mWidth - 1
			Dim argb As ARGBColor
			bc.GetARGB(x, y, argb)
			' Adjust shadows by modifying the darker pixels
			If argb.r < 128 Then argb.r = Min(255, Max(0, argb.r * adjustment))
			If argb.g < 128 Then argb.g = Min(255, Max(0, argb.g * adjustment))
			If argb.b < 128 Then argb.b = Min(255, Max(0, argb.b * adjustment))
			bc.SetARGB(x, y, argb)
		Next
	Next
	Return bc.Bitmap
End Sub

public Sub AdjustBlacks(bmp As B4XBitmap, adjustment As Float) As B4XBitmap
	Dim bc As BitmapCreator = CreateBC(bmp)
	For y = 0 To bc.mHeight - 1
		For x = 0 To bc.mWidth - 1
			Dim argb As ARGBColor
			bc.GetARGB(x, y, argb)
			' Adjust blacks by modifying the very dark pixels
			If argb.r < 64 Then argb.r = Min(255, Max(0, argb.r * adjustment))
			If argb.g < 64 Then argb.g = Min(255, Max(0, argb.g * adjustment))
			If argb.b < 64 Then argb.b = Min(255, Max(0, argb.b * adjustment))
			bc.SetARGB(x, y, argb)
		Next
	Next
	Return bc.Bitmap
End Sub

public Sub AdjustWhites(bmp As B4XBitmap, adjustment As Float) As B4XBitmap
	Dim bc As BitmapCreator = CreateBC(bmp)
	For y = 0 To bc.mHeight - 1
		For x = 0 To bc.mWidth - 1
			Dim argb As ARGBColor 
			bc.GetARGB(x, y, argb)
			' Adjust whites by modifying the very bright pixels
			If argb.r > 192 Then argb.r = Min(255, Max(0, argb.r * adjustment))
			If argb.g > 192 Then argb.g = Min(255, Max(0, argb.g * adjustment))
			If argb.b > 192 Then argb.b = Min(255, Max(0, argb.b * adjustment))
			bc.SetARGB(x, y, argb)
		Next
	Next
	Return bc.Bitmap
End Sub

public Sub AdjustClarity(bmp As B4XBitmap, adjustment As Float) As B4XBitmap
	Dim bc As BitmapCreator = CreateBC(bmp)
	Dim tempBC As BitmapCreator
	tempBC.Initialize(bc.mWidth, bc.mHeight)
    
	' Apply unsharp masking
	Dim radius As Int = 1
	Dim kernelSize As Int = Power((radius * 2 + 1), 2)

	For y = radius To bc.mHeight - radius - 1
		For x = radius To bc.mWidth - radius - 1
			Dim sumR As Int = 0
			Dim sumG As Int = 0
			Dim sumB As Int = 0

			For ky = -radius To radius
				For kx = -radius To radius
					Dim argb As ARGBColor
					bc.GetARGB(x + kx, y + ky, argb)
					sumR = sumR + argb.r
					sumG = sumG + argb.g
					sumB = sumB + argb.b
				Next
			Next

			Dim blurR As Int = sumR / kernelSize
			Dim blurG As Int = sumG / kernelSize
			Dim blurB As Int = sumB / kernelSize
            
			Dim origArgb As ARGBColor
			bc.GetARGB(x, y, origArgb)
			Dim diffR As Int = origArgb.r - blurR
			Dim diffG As Int = origArgb.g - blurG
			Dim diffB As Int = origArgb.b - blurB
            
			origArgb.r = Min(255, Max(0, origArgb.r + diffR * adjustment))
			origArgb.g = Min(255, Max(0, origArgb.g + diffG * adjustment))
			origArgb.b = Min(255, Max(0, origArgb.b + diffB * adjustment))
            
			tempBC.SetARGB(x, y, origArgb)
		Next
	Next
    
	Return tempBC.Bitmap
End Sub

public Sub AdjustVibrance(bmp As B4XBitmap, adjustment As Float) As B4XBitmap
	Dim bc As BitmapCreator = CreateBC(bmp)
    For y = 0 To bc.mHeight - 1
        For x = 0 To bc.mWidth - 1
            Dim argb As ARGBColor
			bc.GetARGB(x, y, argb)
            
            ' Convert RGB to HSL
            Dim hsl As List = RGBtoHSL(argb.r, argb.g, argb.b)
            Dim hue As Float = hsl.Get(0)
            Dim saturation As Float = hsl.Get(1)
            Dim lightness As Float = hsl.Get(2)
            
            ' Adjust vibrance by increasing saturation
            If saturation < 1.0 Then
                saturation = saturation + (1.0 - saturation) * adjustment
                If saturation > 1.0 Then saturation = 1.0
            End If
            
            ' Convert HSL back to RGB
            Dim rgb As List = HSLtoRGB(hue, saturation, lightness)
            argb.r = rgb.Get(0)
            argb.g = rgb.Get(1)
            argb.b = rgb.Get(2)
            
            bc.SetARGB(x, y, argb)
        Next
    Next
	Return bc.Bitmap
End Sub

public Sub RGBtoHSL(r As Int, g As Int, b As Int) As List
    Dim rf As Float = r / 255
    Dim gf As Float = g / 255
    Dim bf As Float = b / 255

    Dim Mx As Float = Max(Max(rf, gf), bf)
    Dim Mn As Float = Min(Min(rf, gf), bf)
    Dim h As Float = 0
    Dim s As Float = 0
    Dim l As Float = (Mx + Mn) / 2
    
    If Mx <> Mn Then
        Dim d As Float = Mx - Mn
        If l > 0.5 Then 
			s = d / (2 - Mx - Mn)
		Else
			s =  d / (Mx + Mn)
		End If
        
        Select Mx
            Case rf
				Dim df As Int 
				If gf < bf Then df = 6 Else df = 0
                h = (gf - bf) / d + df
            Case gf
                h = (bf - rf) / d + 2
            Case bf
                h = (rf - gf) / d + 4
        End Select
        h = h / 6
    End If
    
    Dim result As List
    result.Initialize
    result.Add(h)
    result.Add(s)
    result.Add(l)
    Return result
End Sub

public Sub HSLtoRGB(h As Float, s As Float, l As Float) As List
    Dim r As Float, g As Float, b As Float
    
    If s = 0 Then
        r = l
        g = l
        b = l
    Else
        Dim q As Float 
		If l < 0.5 Then q = l * (1 + s) Else q = l + s - l * s
        Dim p As Float = 2 * l - q
        r = HueToRGB(p, q, h + 1 / 3)
        g = HueToRGB(p, q, h)
        b = HueToRGB(p, q, h - 1 / 3)
    End If
    
    Dim result As List
    result.Initialize
    result.Add(r * 255)
    result.Add(g * 255)
    result.Add(b * 255)
    Return result
End Sub

public Sub HueToRGB(p As Float, q As Float, t As Float) As Float
    If t < 0 Then t = t + 1
    If t > 1 Then t = t - 1
    If t < 1 / 6 Then Return p + (q - p) * 6 * t
    If t < 1 / 2 Then Return q
    If t < 2 / 3 Then Return p + (q - p) * (2 / 3 - t) * 6
    Return p
End Sub

Public Sub CalculateHistogram(bmp As B4XBitmap) As Map
	Dim bc As BitmapCreator = CreateBC(bmp)
	Dim redHistogram(256) As Int
	Dim greenHistogram(256) As Int
	Dim blueHistogram(256) As Int

	For y = 0 To bc.mHeight - 1
		For x = 0 To bc.mWidth - 1
			Dim argb As ARGBColor
			bc.GetARGB(x, y, argb)
			' Increment the histogram values for each color channel
			If argb.r > 0 Then redHistogram(argb.r) = redHistogram(argb.r) + 1
			If argb.g > 0 Then greenHistogram(argb.g) = greenHistogram(argb.g) + 1
			If argb.b > 0 Then blueHistogram(argb.b) = blueHistogram(argb.b) + 1
		Next
	Next

	Dim histograms As Map
	histograms.Initialize
	histograms.Put("red", redHistogram)
	histograms.Put("green", greenHistogram)
	histograms.Put("blue", blueHistogram)
	Return histograms
End Sub



